#!/usr/bin/perl -w
# FIXME 'Chunked' encoding?
#       unfiltered content of unspecified length will cause connection to close
#       (because closed connection signals end of file) -- Chunked encoding can
#       Overcome this.  Note that HTTP::Daemon does not implement it for responses.
# FIXME Uploaded data (forms) broken for large/streaming uploads (who cares?)
# FIXME Large compressable content will be downloaded in its entireity first -- possible
#       client timeout.  Should compress as a stream while feeding to client.
# FIXME Doesn't work for FTP connections.  (that's ok, we're an http proxy)
# FIXME HANG: characterized by "Exiting" message for a process with no corresponding
#       "New connection forked to" message. (rare)

# Header filter: module, call with client's & server's headers.
# Must use HTTP/1.0, and check for Content-Length (no chunked encoding)
# FIXME Server always causes connection to close.
# FIXME If group leader is killed, children aren't, leaving rogue FilterProxy
#       processes.  How to fix this?  (use FilterProxy.pl -k for now)

package FilterProxy;
use vars qw($VERSION $CONFIG @MODULES);
# Place configuration here

my $HOSTNAME        = "localhost";
my $LISTEN_PORT     = 8888;                                   # should be an option
   $VERSION         = 0.17;
my $NAME            = "FilterProxy \$Revision: $VERSION \$";
# NEED ABSOLUTE PATHS!
my $HOME            = `pwd`; # change me!
chomp($HOME);
my $CONFIG_FILE     = $HOME . "/FilterProxy.conf";            # should be an option

# end configuration.

push(@INC, $HOME . "/FilterProxy"); # add my home to perl's include path list
use strict;
no strict 'subs';
no strict 'refs';
use POSIX qw(:sys_wait_h :signal_h setsid setpgid); # part of perl
use Carp;                                           # part of perl
use Config;                                         # part of perl
use File::stat;                                     # part of perl
use CGI;                                            # part of perl
use Getopt::Std;                                    # part of perl
use Data::Dumper;                                   # part of perl
$Data::Dumper::Indent = 1;
use HTTP::Daemon;                                   # in LWP (libwww)
use HTTP::Status;                                   # in LWP
use LWP::UserAgent;                                 # in LWP
use LWP::MediaTypes;                                # in LWP
use Time::HiRes;                                    # grab this from CPAN
use Parse::ePerl;                                   # grab this from CPAN
require Compress::Zlib;                             # grab this from CPAN
# subclass UserAgent to disallow redirects.
@UserAgent::ISA = qw /LWP::UserAgent/;
sub UserAgent::redirect_ok {0;}

# FilterProxy modules (Automate this?)
use FilterProxy::BlockBanner;
#use FilterProxy::Clean;
use FilterProxy::Compress;
use FilterProxy::Skeleton;

# Configuration saved to config file (This is only in case this info isn't
# already in the config file!) 
$CONFIG = {};  # all this is loaded from config file.
$CONFIG->{filtering} = 1; # on/off switch for the filter
$CONFIG->{logfile} = $HOME . "/FilterProxy.log";  # set to "" to disable logfile
$CONFIG->{timeout} = 30;
$CONFIG->{filters} = {};

# Local variables
my(%children) = (); # keys are pids of my children
my($client); # connection to the client
my($inreq);  # incoming request -- needs to be accessable to data_handler
my($grabwholefile); # We're going to grab the whole file, as opposed to passing it through in pieces.
my($received) = 0;
my($length) = 0;
my($last_required) = 0; # cause main loop to read config first time around.
my(@filterable); # list of mime-types that some filter can filter.
my($CR) = "\015"; # Terminates header.
my($LF) = "\012";

# Subroutine prototypes:
sub debug;
sub read_config;
sub write_config;
sub daemonize;        # daemonizes me.
sub handler;          # Handles all incoming requests
sub data_handler;     # Callback when data is received
sub handle_proxy;     # Handles proxy requests
sub handle_filtering; # Takes care of filtering
sub handle_server;    # Handles normal file requests

# Process options
my(%options);
getopts('hkf:p:', \%options);
if($options{'k'}) {  # kill running FilterProxy.
  kill 'INT', `cat $HOME/FilterProxy.pid`;
  exit(0);
}
if($options{'f'}) {
  $CONFIG_FILE = $options{'f'};
}
if($options{'p'}) {
  $LISTEN_PORT = $options{'p'};
}
if($options{'h'} || $ARGV[0]) { # print usage
  print <<END_USAGE;
    Options recognized by FilterProxy:
      -h          Print this help message
      -k          Kill an already running copy of FilterProxy
      -f <file>   Specify an alternate config file 
                    (default is $CONFIG_FILE)
      -p <port>   Specify the port to which FilterProxy will bind 
                    (default is $LISTEN_PORT)
      -n          Do not daemonize: stay connected to the terminal from which
                    it was started and print debugging messages.
END_USAGE
  exit(1);
}

# Read in config file
require $CONFIG_FILE;

# Open log file
if($CONFIG->{logfile} !~ /^\//) {
  $CONFIG->{logfile} = $HOME . "/" . $CONFIG->{logfile};
}
open(LOGFILE, ">>" . $CONFIG->{logfile}) || die "unable to open logfile ", $CONFIG->{logfile}, ".\n";
autoflush LOGFILE 1;
print LOGFILE "\n\n[#######] FilterProxy started. [#######]\n";

# Initialize the daemon.
my $daemon = new HTTP::Daemon LocalAddr => $HOSTNAME, LocalPort => $LISTEN_PORT,
                              Reuse     =>  1
   or croak "HTTP::Daemon failed to initialize: $!\n"
      . "Is $HOSTNAME:$LISTEN_PORT correct?";

# Initialize agent.
my  $agent  =   new UserAgent;
    $agent  ->  agent ($NAME);
    $agent  ->  env_proxy();
    $agent  ->  parse_head(1);
    $agent  ->  timeout($CONFIG->{timeout}); # short timeout -- we retry once.

# perhaps unnecessary, but helpful...
if($CONFIG->{'first_time'}) {
  print "now go to http://$HOSTNAME:$LISTEN_PORT/FilterProxy.html to configure me!\n";
  $CONFIG->{first_time} = 0;
  &write_config;
} else {
  warn "Proxy on ", $daemon->url, " initialized\n";
}

# disassociate from controlling tty, etc, etc.
if(!$options{'n'}) {
  &daemonize; 
}

# Make deaths give line numbers, and wait for children 
sub prefix {
    my $now = localtime;
      join "", map { "[$now] [${$}] $_\n" } split /\n/, join "", @_;
}
$SIG{__WARN__} = sub { warn prefix @_; };
$SIG{__DIE__} = sub { die prefix @_; };

# Zombies are bad.
sub REAPER {
  my($child);
  while (($child = waitpid (-1, &WNOHANG)) && $child != -1) { }
  $SIG{"CHLD"} = \&REAPER;  # still loathe sysV
  delete $children{$child};
}
sub HUNTSMAN {                      # signal handler for SIGINT
    local($SIG{CHLD}) = 'IGNORE';   # we're going to kill our children
    kill 'INT' => keys %children;
    exit;                           # clean up with dignity
}
$SIG {"CHLD"} = \&REAPER;
$SIG {"INT"} = \&HUNTSMAN;
$SIG {"QUIT"} = \&HUNTSMAN;
$SIG {"HUP"} = \&HUNTSMAN;
$SIG {"TERM"} = \&HUNTSMAN;

# And loop forever.
while (my $slave = $daemon -> accept) {
  &read_config;
  if(!defined $slave) { debug("Slave died because: ", $daemon->reason); next; }
  else { 
    defined(my $pid = fork) or die "Can't fork: $!";
    if($pid) { # I am parent
      $children{$pid} = 1;
      close $slave;
    } else { # I am child
      $SIG{INT} = 'DEFAULT';
      handler $slave; 
      exit(1);
    }
  }
}

# That's all folks.
exit 0;

# Blatently ripped from http://www.webreference.com/perl/tutorial/9/3.html, also part of ther perlipc man page...
sub daemonize {
  chdir '/'                  or die "Can't chdir to /: $!";
  defined(my $pid = fork)    or die "Can't fork: $!";
  exit if $pid;
  setsid                     or die "Can't start a new session: $!";
  setpgrp; # don't think I need both...
# can only do this as root?
# setpgid(0,0)               or die "Can't set group ID: $!";
  open STDIN, '/dev/null'    or die "Can't read /dev/null: $!";
  open STDOUT, '>>/dev/null' or die "Can't write to /dev/null: $!";
  open STDERR, '>>/dev/null' or die "Can't write to /dev/null: $!";
  umask 0;
  open(PIDFILE, ">$HOME/FilterProxy.pid"); # so I can kill myself easily
  print PIDFILE $$;
  close(PIDFILE);
}

sub debug {
  if($_[$#_] !~ /\n$/) { push @_, "\n"; }
  my $now = localtime;
  if(defined LOGFILE) {
    print LOGFILE "[$$ $now] ", @_;
  }
  if($CONFIG->{debug}) { # doesn't work when daemonized, cuz STDOUT -> /dev/null
    print "[$$ $now] ", @_;
  }
}

sub read_config {
  # Reconfigure if the config file changed.
  my $mod = (stat $CONFIG_FILE)->mtime || 0;
  if ($mod > $last_required) {
    delete $INC {$CONFIG_FILE}; # tell perl we haven't already require'd it.
    require $CONFIG_FILE;
    $last_required = $mod;
    my($module); # make a list of mime-types that some filter can filter.
    @filterable = ();
    foreach $module (@MODULES) {
      push @filterable, @{${"FilterProxy::" . $module . "::CONFIG"}->{mime_types}};
      my(%filterable); # uniqueify list.
      foreach my $mime_type (@filterable) { $filterable{$mime_type} = 1; }
      @filterable = sort keys %filterable;
    }
  }
}

sub write_config {
  my($config, $module);
  open(CONF, ">$CONFIG_FILE") || die "unable to open config file.\n";
  print CONF "# FilterProxy Configuration file.\n";
  print CONF "#\n";
  print CONF "# Be aware that FilterProxy will automatically rewrite this file.\n";
  print CONF "# It will also automatically reload this file if it detects that it\n";
  print CONF "# has been modified, so there is no need to kill FilterProxy when\n";
  print CONF "# editing by hand.\n";
  print CONF "#\n";
  print CONF "# This is a normal perl file, and will be eval'ed by FilterProxy\n";
  print CONF "# to load it's configuration.  I recommend that initially, you use\n";
  print CONF "# the forms interface to add stuff, and then look at this file to\n";
  print CONF "# see its structure.\n";

  foreach $module ("FilterProxy", map("FilterProxy::" . $_, @MODULES)) {
    $config = $module . "::CONFIG";
    print CONF "\n" . Data::Dumper->Dump([$$config], ["$config"]);
  }
  print CONF "1;\n";
  close CONF;
}

sub handler {
  $client = shift; # connection to the client
  $received = 0;
  $length = 0;
  my($key,$val);
  my($requests) = 0;
  my($start, $end);
  $client->timeout(300); # If we don't see a request within 5 minutes, bail.
  while($client) {
    $inreq = $client->get_request(0); # 1=only read headers. 0=read whole request
    if(!defined $inreq) { 
      $client->close;
      undef($client);
      debug("Exiting. ($requests requests processed)\n");
      exit(0);
    }
    &read_config;
    $start = Time::HiRes::time;
    $requests++;
    if($inreq->uri !~ /http:\/\/$HOSTNAME:$LISTEN_PORT/) {
      &handle_proxy($inreq, $client);
    } else { # Server mode -- serve up configuration pages.
      &handle_server($inreq, $client);
    }
    $end = Time::HiRes::time;
    debug("Request took ", $end-$start, " seconds total.\n");
  }
  debug("EXITING IN A BAD PLACE!!!!\n");
  exit(1);
}

sub data_handler {
  my($data, $res, $protocol) = @_;
  if($received == 0) { # if this is the first data block, do headers
    if($inreq->uri !~ /http:\/\/$HOSTNAME:$LISTEN_PORT/ and defined $inreq->header('Proxy-Connection')) {
      $res->remove_header('Connection');
      $res->remove_header('Xonnection');
      $res->header('Proxy-Connection' => $inreq->header('Proxy-Connection'));
    }
    if(!defined $res->content_length || grep {$res->content_type =~ $_} @filterable) {
      $grabwholefile = 1;
    }
    if(!$grabwholefile) {
      $client->send_status_line($res->code, $res->message);
      print $client $res->headers_as_string . "$CR$LF";
    }
    $length = $res->content_length?$res->content_length:-1;
  }
  $received += length $data;
  if($grabwholefile) {
    $res->add_content($data);   # stow data, filter at end
  } else {
    print $client $data;
  }
}

# Handle one proxy request.
sub handle_proxy {
  my($origreq) = shift;
  my($req) = $origreq->clone;
#debug($req->headers_as_string);
# FIXME filter headers here order: -1..-3

# This is a hack around HTTP::Daemon not recognizing Proxy-Connection headers.
  if(defined $req->header('Proxy-Connection') && $req->header('Proxy-Connection') =~ /\bkeep-alive\b/i) {
    ${*$client}{'httpd_nomore'} = 0;
  }
# Can't handle keep-alive proxy connections (yet).  (LWP::UserAgent limitation)
  if(defined $ENV{http_proxy}) { 
    $req->header('Proxy-Connection', 'close'); 
  } else {
    $req->remove_header('Proxy-Connection');
  }

# uploaded content
  if(defined $req->content_length && $req->content_length > 0) {
# FIXME fix for streaming uploaded content.
#    $req->read($content, $req->header('Content-length'));
  }
  $received = 0;
  $grabwholefile = 0; # set by data_handler from header data
  debug("Sending request for ", $req->uri, "\n");
  my $res = $agent->request($req, \&data_handler);
  debug("[", $res->code, " ", $res->message, "] for ", $inreq->uri, 
      ($res->content||$res->code == 200)
        ?join("", " (", $res->content_type, " ", 
          ($res->content_length?$res->content_length:"unspecified"), " bytes)\n")
        :""
      );
  if(defined $origreq->header('Proxy-Connection')) { # duplicated effort if there is content (data_handler called)
    $res->header('Proxy-Connection' => $origreq->header('Proxy-Connection'));
  }
  if($grabwholefile || $res->is_redirect) {
    my($content_type) = $res->content_type;
    if(scalar(grep {$res->content_type =~ $_} @filterable) && $CONFIG->{filtering}) {
      if(defined $res->content_encoding) {
        debug("Found Content-Encoding: ", $res->content_encoding, "\n");
        if($res->content_encoding =~ /gzip/i) {

# FIXME This is a blatent hack.  None of the other methods here work, and I don't know why.

#          my($test) = Compress::Zlib::uncompress($res->content);
#          if(!defined $test) { debug("test is undefined.\n"); }
#          my($inflater, $status) = Compress::Zlib::inflateInit();
#          if(!defined $inflater) { debug("Zlib init failed: status is $status\n"); }
#          ($test, $status) = $inflater->inflate($res->content);
#          if(!defined $test) {
#            debug("Error uncompressing gzip encoded response! $status\n");
          eval {
            open(GZIPSAVE, ">/tmp/gzipsave.gz") || die "error opening GZIPSAVE.\n";
            print GZIPSAVE $res->content;
            close GZIPSAVE;
            my($gz) = Compress::Zlib::gzopen("/tmp/gzipsave.gz", "rb") || die("Cannot open gzip file.\n");
            my($buffer, $content) = ("", "");
            $content .= $buffer while $gz->gzread($buffer) > 0 ;
            if($gz->gzerror !~ /stream end/) { die "GZ error: ", $gz->gzerror; }
            $res->content($content);
            $res->remove_header('Content-Encoding');
          }
        }
      }
      &handle_filtering($res);
    }
    if(length ${$res->content_ref} == 0) { # send an empty file
      $res->content(" ");
    }
    $client->send_status_line($res->code, $res->message);
    $res->content_length(length ${$res->content_ref});
# FIXME filter response headers here (order = 10)
    print $client $res->headers_as_string . "$CR$LF";
    print $client ${$res->content_ref};
  } elsif($res->is_error) {
    $client->send_status_line($res->code, $res->message);
    if(!(defined $res->content_ref && length ${$res->content_ref} != 0)) {
      $res->content($res->error_as_HTML);
    }
    $res->content_length(length ${$res->content_ref});
    print $client $res->headers_as_string . "$CR$LF";
    print $client ${$res->content_ref};
  } elsif($res->code != 200) {
    debug("(Unhandled response!) What kind of response is [", $res->code, " ", $res->message, "]\n");
  }
  undef($res); # kill the response and all its data
}

sub handle_filtering {
  my($res) = shift;
  if ($res->is_success) {
#   Do filtering.
    my(%filterconfig);
    my($regexp, $filter, $mime_type);
REGEXP:
    foreach $regexp (keys(%{$CONFIG->{filters}})) {
      next REGEXP unless ($inreq->uri =~ $regexp);
FILTER:
      foreach $filter (keys(%{$CONFIG->{filters}->{$regexp}})) {
        next FILTER unless scalar(grep {$res->content_type =~ $_} 
            @{${"FilterProxy::" . $filter . "::CONFIG"}->{mime_types}});
        next unless defined &{"FilterProxy::" . $filter . "::filter"};
        # prepend the order here so that sort (below) will put things in order.
        push @{$filterconfig{${"FilterProxy::" . $filter . "::CONFIG"}->{order} . $filter}}, 
          $CONFIG->{filters}->{$regexp}->{$filter};
        debug("Going to apply filter $filter for ", $inreq->uri, ", matching /$regexp/.\n");
      }
    }
    foreach $filter (sort keys %filterconfig) {
      my ($start, $stop);
      my ($len) = length ${$res->content_ref};
      my($basefilter) = $filter;
      $basefilter =~ s/^\d+//; # remove preceeding order
      $start = Time::HiRes::time();
      &{"FilterProxy::" . $basefilter . "::filter"}($inreq, $res, $filterconfig{$filter}); 
      $stop = Time::HiRes::time();
      my $rate = $len/($stop-$start);
      debug("  $basefilter took ", $stop-$start, " seconds to process ", $len,
          " bytes: ", $rate, " bytes/sec\n");
    }
  } else {
    debug($res->request->uri, " not filtered because ", $res->code, " ", $res->message, "\n");
  }
}

sub handle_server {
  my($origreq, $client) = @_;
  my($req) = $origreq->clone;
  my($params);
# This is a hack around HTTP::Daemon not recognizing Proxy-Connection headers.
  if(defined $req->header('Proxy-Connection') && $req->header('Proxy-Connection') =~ /\bkeep-alive\b/i) {
    ${*$client}{'httpd_nomore'} = 0;
  }
# Can't handle keep-alive proxy connections (yet).  (LWP::UserAgent limitation)
  if(defined $ENV{http_proxy}) { 
    $req->header('Proxy-Connection', 'close'); 
  } else {
    $req->remove_header('Proxy-Connection');
  }

  if($req->uri =~ /^.*?\?(.*)$/) {
    if($req->content) { $params = $req->content . "&$1"; }
    else { $params = $1; }
  } else {
    $params = $req->content_ref?$req->content:"";
  }
  my($cgi) = new CGI($params);
  my($file, $module);
  my($message) = "";
  my($SITECONFIG, $MODULECONFIG, $SITE);
  local(*FILE);
  my($res) = new HTTP::Response(RC_OK); # need no-cache

  debug("Serving request for ", $req->uri);
  $res->server("$HOSTNAME:$LISTEN_PORT");
  $res->date(time);
  if($req->uri =~ /^http:\/\/$HOSTNAME:$LISTEN_PORT\/(.*?)(\?.*)?$/) {
    $file = "$1";
    if(-d "$HOME/html/$file") { # serve index.html, just like a regular web server
      $file .= "/index.html";
    }
    if(-f "$HOME/html/$file") {
      guess_media_type("$HOME/html/$file", $res); # sets content-type
      my($stat) = stat "$HOME/html/$file";
      my($size, $mtime) = ($stat->size, $stat->mtime);
      $res->last_modified($mtime) if $mtime;
      if($res->content_type =~ /text\/html/ || $req->uri =~ /\.html$/) { # pass the mofo through Parse::ePerl::Expand
        if($file =~ /^([a-zA-Z]+)\.?.*?\.html$/) {
            $module = $1;
        } else {
          debug("Unable to determine module for request!");
        }
        my(@mods) = grep {/$module$/} (map("FilterProxy::" . $_, @MODULES), "FilterProxy");
        my($fullmodule);
        if($#mods == 0) {
          $fullmodule = $mods[0];
          $MODULECONFIG = ${$fullmodule . "::CONFIG"};
          if($module !~ /FilterProxy/ && $cgi->param('site') && 
              defined $CONFIG->{filters}->{$cgi->param('site')}) { 
            if(!defined $CONFIG->{filters}->{$cgi->param('site')}->{$module}) {
              $CONFIG->{filters}->{$cgi->param('site')}->{$module} = {}; 
            }
            $SITECONFIG = $CONFIG->{filters}->{$cgi->param('site')}->{$module}; 
            $SITE = $cgi->param('site');
          } else {
            $SITECONFIG = undef;
            $SITE = undef;
          }
          $message = &{$fullmodule . "::Config"}($req, $cgi, $SITECONFIG);
          &write_config(); # save changes.
        } else {
          debug("Unrecognized module: $module");
        }
        if(!open(FILE, "<$HOME/html/$file")) { 
          debug("unable to open $file"); 
          $client->send_error(RC_FORBIDDEN); 
          return; 
        }
        my($script) = join("", <FILE>);
        close(FILE);
        my($content, $error);
        my(%params);
        if(defined $MODULECONFIG) { $params{MODULECONFIG} = $MODULECONFIG; };
        if(defined $SITECONFIG) { $params{SITECONFIG} = $SITECONFIG; };
        if(defined $SITE) { $params{SITE} = $SITE; }
        $params{VERSION} = defined ${$fullmodule . "::VERSION"}?${$fullmodule . "::VERSION"}:"";
        $params{MESSAGE} = defined $message?$message:"";
        $params{MODULES} = \@MODULES; # properly, this should be in FilterProxy::Config
        my($result) = Parse::ePerl::Expand({Script => $script, Name => $file,
            Result => \$content, Error => \$error, ENV => \%params});
        if(!$result) { $client->send_error(RC_INTERNAL_SERVER_ERROR); debug("ePerl error: $error"); return; }
        if($content) { $res->content($content); 
        }
        $res->header('Pragma' => 'no-cache');
        $res->content_length(length(${$res->content_ref}));
      } else {
        $res->content_length($size) if $size;
        sysopen(FILE, "$HOME/html/$file", 0) || $client->send_error(RC_FORBIDDEN);
        $res->content(undef);
      }
    } else { $client->send_error(RC_NOT_FOUND); return; }
  } else { $client->send_error(RC_BAD_REQUEST); return; }

  if(defined $origreq->header('Proxy-Connection')) { # duplicated effort if there is content (data_handler called)
    $res->header('Proxy-Connection' => $origreq->header('Proxy-Connection'));
  }
  $client->send_status_line($res->code, $res->message);
  print $client $res->headers_as_string . "$CR$LF";
  if(defined $res->content) {
    print $client $res->content;
  } else {
    $client->send_file(\*FILE);
    close(FILE);
  }
  undef($res);
}

sub Config { 
# takes request, cgi, returns HTML message.
  my($req, $cgi, $config) = @_;
  my($site);
  my($module);
  my($message) = ""; 
# html message describing success/failure for the client.

  if($cgi->param()) {
    debug("parameters: ", join(" ", $cgi->param()), "\n");
    if(defined $cgi->param('logfile') && defined $cgi->param('timeout')) {
      if(defined $cgi->param('filtering')) { $CONFIG->{filtering} = 1; }
      else { $CONFIG->{filtering} = 0; }
      if(defined $cgi->param('debug')) { $CONFIG->{debug} = 1; }
      else { $CONFIG->{debug} = 0; }
      $CONFIG->{timeout} = $cgi->param('timeout');
      $CONFIG->{logfile} = $cgi->param('logfile');
    }
    if($cgi->param("add") || (defined $cgi->param("add.x")) || (defined $cgi->param("add.y"))) {
      if($cgi->param('url') and ($cgi->param('url') !~ $cgi->param('newsite'))) {
        $message = "<center><font color=red><h3>Your selected URL regexp, \"" . 
            $cgi->param('newsite') . "\"does not match the URL you selected." . 
            "</font></center>\n";
      } else {
        debug("site: ", $cgi->param('site'), "\n");
        debug("module: ", $cgi->param('module'), "\n");
        if(!defined $CONFIG->{filters}->{$cgi->param('site')}) {
          $CONFIG->{filters}->{$cgi->param('site')} = {};
        }
        if(!defined $CONFIG->{filters}->{$cgi->param('site')}->{$cgi->param('module')}) {
          $CONFIG->{filters}->{$cgi->param('site')}->{$cgi->param('module')} = {};
        }
        $message = "<center><font color=blue>Filter module " . 
            $cgi->param('module') . " successfully added for site regexp " . 
            $cgi->param('site') . ".</font></center>\n";
      }
    }
    if($cgi->param('delete') || $cgi->param('delete.x') || $cgi->param('delete.y')) {
      if($cgi->param('module') && $cgi->param('site')) {
        debug("Deleting module ", join(",", $cgi->param('module')), " from site ", join(",", $cgi->param('site')));
        debug(Data::Dumper->Dump([$CONFIG->{filters}->{$cgi->param('site')}], [qw($CONFIG->{filters}->{$cgi->param('site')})]));
        delete $CONFIG->{filters}->{$cgi->param('site')}->{$cgi->param('module')};
        my(@keys) = keys %{$CONFIG->{filters}->{$cgi->param('site')}};
        if($#keys == -1) {
          delete $CONFIG->{filters}->{$cgi->param('site')};
        }
        debug(Data::Dumper->Dump([$CONFIG->{filters}->{$cgi->param('site')}], [qw($CONFIG->{filters}->{$cgi->param('site')})]));
      } elsif($cgi->param('site')) {
        debug("Deleting site ", $cgi->param('site'), "\n");
        delete $CONFIG->{filters}->{$cgi->param('site')};
      }
    }
  }
  return $message;
}

1;

__END__

