# Skeleton FilterProxy Module

# This is an example module for those interested in writing new modules.
# It simply adds a string to the beginning of a web page, nothing more.

package FilterProxy::Skeleton;

use strict;
use vars qw($VERSION $CONFIG);  # The $CONFIG variable MUST be exported.
                                # and should be a reference to a hash.
                                # keys to this hash are configuration
                                # variables, and values are used by this 
                                # module.

push @FilterProxy::MODULES,     # add myself to FilterProxy's list of known 
   "Skeleton";                  # modules
$VERSION = 0.01;
$CONFIG = {};                   # my $CONFIG.  This will be saved to and
                                # loaded from the config file by FilterProxy
                                # Module should define a basic CONFIG just in
                                # case this is the first time this module
                                # has ever been used.

$CONFIG->{order} = 4;           # must be a value in the range 0-10
                                # which specifies where in the chain of
                                # filters it wants to be executed.
                                # < 0: called before request is sent
                                # >= 0: called after content is received
                                # -3--1: Must be called before request is sent
                                #   (may munge request's headers, encoding, etc.)
                                # 0: Must be called before anyone touches the
                                #   content, and before headers are munged. (content decoders, debug)
                                # 1-3: May munge headers, will not touch content.
                                # 4-6: Will modify content (content filters)
                                # 7-9: May encode content (i.e. content will
                                #   be unreadable to filters in 4-6).
                                # 10: Must be called after all filters. (final header manipulation)
                                # This may also be an anonymous list ref if
                                # the module needs to be called more than once
                                # for a given request. (i.e. header mungers)

$CONFIG->{mime_types} = [       # is a list of mime-types that this module
  "text/html",                  # knows how to filter.  Can contain an empty
  "text/plain",                 # string to specify all mime-types.
];

*debug = \&FilterProxy::debug;  # Tells perl to use FilterProxy's debug function
                                # instead of looking for 
                                # FilterProxy::Skeleton::debug (this is because
                                # I'm lazy and don't want to type 
                                # FilterProxy::debug every time)
                                # Using this function will cause data to go to 
                                # FilterProxy's log file.

sub filter {                    # The main filtering routine.  Does the dirty work.
  my($req) = shift;             # $req is the HTTP::Request object 
  my($res) = shift;             # $res is the HTTP::Response object.  
  my($siteconfig) = shift;      # $siteconfig is a reference to a list of
                                # the site-specific configuration(s) for the
                                # site(s).  non site-specific config should 
                                # be in the
                                # $CONFIG variable (see above).

                                # Module must be able to handle it if any of 
                                # the above parameters are undef.
  if(defined $res) {
    ${$res->content_ref} = 
      "<h1><center>FilterProxy::Skeleton modified this file</center></h1>"
      . ${$res->content_ref};
  }
  debug("FilterProxy::Skeleton::filter called.");
                                # $res->content_ref is an efficent way of
                                # accessing the document. (See HTTP::Message
                                # man page).
  return;                       # nothing needs to be returned.
}

sub Config {                    # parses the CGI parameters when the user
                                # modifies the configuration for this module.
                                # There should also be a .html file with the
                                # same name as the module, which contains the
                                # forms processed by this CGI.
  my($req, $cgi, $siteconfig) = @_;

  return "It worked.\n";        # a HTML-formatted message that will be 
                                # available to your .html file as $ENV{MESSAGE}
}

# NOTES:
#
# Is it advantageous to make FilterProxy::Module, and make everything an object?
